% Applies the Linear Discriminant Analysis algorythm for the already seen
% samples in cluster with id c_id and splits the seen and unseen ones in 
% new clusters. 
%
% IN
% c_id: id of the cluster
%
% OUT
% elems: Ids for all the elements in the cluster
% cluster: Id of the cluster that each elem belongs to
% original_features: original features from this elements
function [elems cluster original_features] = apply_LDA(c_id, nClusters)

    global C;
    global PCAthreshold;
    global LDAthreshold;

    % Gets the cluster idx
    found = false;
    i = 1;
    while(~found)
        if(C(i).id == c_id)
            found = true;
            c_id = i;
        end
        i = i+1;
    end
    


    %% We get the seen elements
    seen_elems = C(c_id).seen_idx;
    lenSeen = length(seen_elems);
    seen_elems_pos = [];
    for i = 1:lenSeen
        seen_elems_pos = [seen_elems_pos find(C(c_id).elements_id==seen_elems(i))];
    end
    % We get all the features of the seen elements
    seen_features = C(c_id).features(seen_elems_pos, :);
    seen_labels = C(c_id).seen_labels;
    uniqueLabels = unique(seen_labels);
    
    
    %% We get the unseen elements
    unseen_elems = setdiff(C(c_id).elements_id, C(c_id).seen_idx);
    lenUnseen = length(unseen_elems);
    unseen_elems_pos = [];
    for i = 1:lenUnseen
        unseen_elems_pos = [unseen_elems_pos find(C(c_id).elements_id==unseen_elems(i))];
    end
    % We get the features of the unseen elements
    unseen_features = C(c_id).features(unseen_elems_pos, :);
    
    
    %% KNN classification for the unseen samples
    unseen_labels = KNN_for_tree(uniqueLabels, seen_labels, seen_features, unseen_features);
    
    
    %% Balancing for PCA
    all_features = [seen_features; unseen_features];
    all_labels = [seen_labels; unseen_labels];
    
    count = zeros(length(uniqueLabels), 1);
    posElems = {};
    for i = 1:length(uniqueLabels)
        posElems{i} = [];
    end
    for i = 1:(lenSeen+lenUnseen)
        pos = find(uniqueLabels==all_labels(i));
        count(pos) = count(pos) +1;
        posElems{pos} = [posElems{pos} i];
    end
    numPicks = min(count);
    chosen_features = [];
    chosen_labels = [];
    for i = 1:length(uniqueLabels)
        % Picks the random indx to train
        M = posElems{i};
        p = randperm(length(M)); % Reshuffle their order randomly
        chosen_features = [chosen_features; all_features(M(p(1:numPicks)), :)]; % Choose the first "numPicks" of these
        chosen_labels = [chosen_labels; all_labels(M(p(1:numPicks)))];
    end

    %% PCA: Calculates covariance matrix, eigenvalues and eigenvectors
    Cov = cov(chosen_features);
    [V_PCA, D_PCA] = eig(Cov); % V = eigenvectors, D = eigenvalues

    % Calculates the features in the new space and deletes the unuseful ones
    vectors = [];
    for i = 1:size(D_PCA,1)
        if(D_PCA(i,i) > PCAthreshold)
            vectors = [vectors; i];
        end
    end
    V_PCA = V_PCA(:,vectors);
    original_features = [seen_features; unseen_features];
    media = mean([seen_features; unseen_features]);
    seen_features = (seen_features - repmat(media, size(seen_features,1),1)) * V_PCA;
    unseen_features = (unseen_features - repmat(media, size(unseen_features,1),1)) * V_PCA;
    
    
    %% We balance the number of elements of each class for the LDA
    all_features = [seen_features; unseen_features];
    all_labels = [seen_labels; unseen_labels];
    
    count = zeros(length(uniqueLabels), 1);
    posElems = {};
    for i = 1:length(uniqueLabels)
        posElems{i} = [];
    end
    for i = 1:(lenSeen+lenUnseen)
        pos = find(uniqueLabels==all_labels(i));
        count(pos) = count(pos) +1;
        posElems{pos} = [posElems{pos} i];
    end
    numPicks = min(count);
    chosen_features = [];
    chosen_labels = [];
    for i = 1:length(uniqueLabels)
        % Picks the random indx to train
        M = posElems{i};
        p = randperm(length(M)); % Reshuffle their order randomly
        chosen_features = [chosen_features; all_features(M(p(1:numPicks)), :)]; % Choose the first "numPicks" of these
        chosen_labels = [chosen_labels; all_labels(M(p(1:numPicks)))];
    end

    
    %% LDA: Calculate linear discriminant analysis

    % We join all the elements to merge them in different clusters
    elems = [seen_elems unseen_elems];
    elems_features = [seen_features; unseen_features];
    
    % Transforms the features using the between-class and within-class scatter matrices
    [Sb, Sw] = scattermat(chosen_features, chosen_labels);
    [V_LDA, D_LDA] = eig(Sb,Sw); % V = eigenvectors, D = eigenvalues
    
    vectors = [];
    for i = 1:size(D_LDA,1)
        if(D_LDA(i,i) > LDAthreshold)
            vectors = [vectors; i];
        end
    end
    V_LDA = V_LDA(:,vectors);
    
    transformed_features = elems_features * V_LDA;
    

    %% KMeans clustering
    if(~isempty(transformed_features))  
        [cluster, centroid] = kmeans(transformed_features, nClusters, 'emptyaction', 'singleton');
    else
        cluster = all_labels;
        % If any of them was semiplate we put it in the plate group
        if(nClusters < 3)
            cluster(find(cluster==3)) = 2;
        end
    end


end


%% Calculates the between and within scatter matrices
function [B W] = scattermat(data,Y)
%FUNCTION THAT CALCULATES SCATTER MATRIX:
%   B:BETWEEN  CLASS SCATTER MATRIX
%   W:WITHIN CLASS SCATTER MATRIX
%

    [~, l]=size(data);         %CALCULATE SIZE OF DATA
    clases=unique(Y);          %GET VECTOR OF CLASSES
    tot_clases=length(clases); %HOW MANY CLASSES
    B=zeros(l,l);              %INIT B AND W
    W=zeros(l,l);           
    overallmean=mean(data);    %MEAN OVER ALL DATA
    for i=1:tot_clases
        clasei = find(Y==clases(i)); %GET DATA FOR EACH CLASS
        xi=data(clasei,:);
        
        mci=mean(xi);                       %MEAN PER CLASS
        xi=xi-repmat(mci,length(clasei),1); %Xi-MeanXi
        W=W+xi'*xi;                         %CALCULATE W
        B=B+length(clasei)*(mci-overallmean)'*(mci-overallmean); %CALCULATE B
    end
end

%     % Class' features distribution Figure
%     fig = figure;
%     for i = 1: length(seen_features)
%         if(seen_labels(i) == 1)
%             plot(seen_features(i,:), 'red');
%         elseif(seen_labels(i) == 2)
%             plot(seen_features(i,:), 'green');
%         else
%             plot(seen_features(i,:), 'yellow');
%         end
%         hold on;
%     end
%     legend({'No Plate'; 'Plate'; 'Semiplate'});
%     title('Class features distribution');
%     saveas(fig,'samples_distribution', 'jpg');


% %     Classes separated with LDA
%     fig = figure;
%     gscatter(transformed_features(1:size(seen_features,1),1), transformed_features(1:size(seen_features,1),2), seen_labels, 'rgy','ooo');
%     legend('No Plate', 'Plate', 'Semiplate');
%     title('Samples in LDA space');
%     
%     saveas(fig, 'Samples in LDA space', 'jpg');
% 
% 
% %     Clustering of samples in LDA space
%     fig2 = figure;
%     c = {};
%     lbls = [];
%     comp1 = [];
%     comp2 = [];
%     lgnd = {};
%     for i = 1:nClusters
%         c{i} = find(cluster == i);
%         posi = find(c{i} <= size(seen_features,1));
%         c{i} =c{i}(posi);
%         lbls = [lbls; ones(length(c{i}),1)*i];
%         comp1 = [comp1; transformed_features(c{i},1)];
%         comp2 = [comp2; transformed_features(c{i},2)];
%         lgnd{i} = ['Cluster ' num2str(i)];
%     end
%     gscatter(comp1, comp2, lbls, 'mbrgy','o');
%     legend(lgnd);
%     title('Clustering of samples in LDA space');
%     
%     saveas(fig2, 'Clustering of samples in LDA space', 'jpg');
